import numpy as np
from scipy.optimize import broyden1
from scipy.optimize import nonlin

# tube triode model
# Also has method for junction FET
# Default values are typical of a 12AX7
class tMod():
    def __init__(self, mu = 100., X = 1.4, KG1 = 650., kp = 600., kVB = 300.,
         RL = 1.e5, Epmax = 400., Vsup = 300., IDSS = .003, Vpinch = 3.0):
        self.mu = mu
        self.X = X
        self.KG1 = KG1
        self.kp = kp
        self.kVB = kVB
        self.RL = RL
        self.Epmax = Epmax
        self.Vsup = Vsup
        self.IDSS = IDSS
        self.Vpinch = Vpinch
        self.Ep = np.linspace(0., Epmax, num=100)
        self.angle = np.linspace(0., 360., num= 256, endpoint=False)

# Simple model: For finding the plate current uisng the defined Ep and a supplied EG
    def mIp(self, EG):
        return (np.clip(self.Ep/self.mu + EG, 0., 1.e33))**1.5/self.KG1

# Simple model equation for use in solver (broyden1).
    def Fsim(self, Ep):
        return np.clip((Ep/self.mu + self.EG)**1.5, 0., 1.e33)/self.KG1 - (self.Vsup - Ep)/self.RL

# Solves for Ep when given an Eg and the parameters set up above
# The solution fails if an attempt is made to go above the power supply voltage,
# and so the power supply voltage is returned.
# Uses simple model
    def ss(self, EG, guess = 400., f_tol = 1.e-14):
        self.EG = EG
        try:
            return broyden1(self.Fsim, guess, f_tol = f_tol)
        except nonlin.NoConvergence as e:
            print('Did not converge:{0}'.format(e))
            return self.Vsup

# Takes a bias EG and a peak Eg for the added sine wave variation
# and makes 256 values of Ep over one sine wave cycle.
# Uses simple model
    def sResp(self, EGb, EGs):
        result = np.zeros(256)
        wt = np.linspace(0., 2.*np.pi, num = 256, endpoint = False)
        EGs = EGb + EGs*np.sin(wt)
        for i in range(result.size):
            result[i] = self.ss(EGs[i])
        return result

# New model: For finding the plate current uisng the defined Ep and a supplied EG
# http://gabevee.tripod.com/biasing.html
    def mIpn(self, EG):
        self.E1 = (self.Ep/self.kp)*np.log(1. \
            + np.exp(self.kp*(1./self.mu + EG/np.sqrt(self.kVB + self.Ep**2))))
        return (np.clip(self.E1, 0., 1.e33))**1.5/self.KG1

# New model equation for use in solver (broyden1).
    def Fsimn(self, Ep):
        E1 = (Ep/self.kp)*np.log(1. \
            + np.exp(self.kp*(1./self.mu + self.EG/np.sqrt(self.kVB + Ep**2))))
        return np.clip((E1)**1.5, 0., 1.e33)/self.KG1 - (self.Vsup - Ep)/self.RL

# Solves for Ep when given an Eg and the parameters set up above
# The solution fails if an attempt is made to go above the power supply voltage,
# and so the power supply voltage is returned.
# Uses new model
    def ssn(self, EG, guess = 400., f_tol = 1.e-14):
        self.EG = EG
        try:
            return broyden1(self.Fsimn, guess, f_tol = f_tol)
        except nonlin.NoConvergence as e:
            print('Did not converge:{0}'.format(e))
            return self.Vsup

# Takes a bias EG and a peak Eg for the added sine wave variation
# and makes 256 values of Ep over one sine wave cycle.
# Uses new model
    def sRespn(self, EGb, EGs):
        result = np.zeros(256)
        wt = np.linspace(0., 2.*np.pi, num = 256, endpoint = False)
        EGs = EGb + EGs*np.sin(wt)
        for i in range(result.size):
            result[i] = self.ssn(EGs[i])
        return result

# FET response. 
    def FResp(self, EGb, EGs):
        wt = np.linspace(0., 2.*np.pi, num = 256, endpoint = False)
        EGs = EGb - EGs*np.sin(wt)
        ID = self.IDSS*(1 - EGs/self.Vpinch)**2
        return ID


